import pytest

from random_order.shuffler import _shuffle_items


def identity_key(item, session):
    return item


def modulus_2_key(item, session):
    return item % 2


def lt_10_key(item, session):
    return item < 10


def disable_if_gt_1000(item, session):
    # if disable returns a truthy value, it must also be usable as a key.
    if item > 1000:
        return item // 1000
    else:
        return False


@pytest.mark.parametrize(
    "key",
    [
        None,
        lambda item, session: None,
        lambda item, session: item % 2,
    ],
)
def test_shuffles_empty_list_in_place(key):
    items = []
    items_id = id(items)
    _shuffle_items(items, bucket_key=key)
    assert items == []
    assert id(items) == items_id


@pytest.mark.parametrize(
    "key",
    [
        None,
        lambda item, session: None,
        lambda item, session: item % 2,
    ],
)
def test_shuffles_one_item_list_in_place(key):
    items = [42]
    items_id = id(items)
    _shuffle_items(items, bucket_key=key)
    assert items == [42]
    assert id(items) == items_id


def test_identity_key_results_in_complete_reshuffle():
    items = list(range(20))
    items_copy = list(items)
    _shuffle_items(items, bucket_key=identity_key)
    assert items != items_copy


def test_two_bucket_reshuffle():
    # If we separate a list of 20 integers from 0 to 19 into two buckets
    # [[0..9], [10..19]] then the result should be
    # either [{0..9}, {10..19}]
    # or [{10..19}, {0..9}]
    items = list(range(20))
    items_copy = list(items)
    _shuffle_items(items, bucket_key=lt_10_key)
    assert items != items_copy
    for i, item in enumerate(items):
        if lt_10_key(i, None):
            assert lt_10_key(item, None) == lt_10_key(items[0], None), items
        else:
            assert lt_10_key(item, None) == lt_10_key(items[10], None), items


def test_eight_bucket_reshuffle():
    # This is a cross-check to test shuffling of buckets.
    items = [
        1,
        1,
        2,
        2,
        3,
        3,
        4,
        4,
        5,
        5,
        6,
        6,
        7,
        7,
        8,
        8,
    ]
    items_copy = list(items)

    _shuffle_items(items, bucket_key=identity_key)

    assert items != items_copy

    # Items of the same bucket should remain next to each other.
    assert items[0] == items[1]
    assert items[2] == items[3]
    assert items[-1] == items[-2]


def test_shuffle_respects_single_disabled_group_in_each_of_two_buckets():
    items = [
        11,
        13,
        9995,
        9997,
        19,
        21,
        23,
        25,
        27,
        29,  # bucket 1 -- odd numbers
        12,
        14,
        9996,
        9998,
        20,
        22,
        24,
        26,
        28,
        30,  # bucket 2 -- even numbers
    ]
    items_copy = list(items)

    _shuffle_items(items, bucket_key=modulus_2_key, disable=disable_if_gt_1000)

    assert items != items_copy
    assert items.index(9995) + 1 == items.index(9997)
    assert items.index(9996) + 1 == items.index(9998)


def test_shuffle_respects_two_distinct_disabled_groups_in_one_bucket():
    # all items are in one oddity bucket, but the two groups
    # of large numbers are separate because they are disabled
    # from two different units.
    # This is simulating two disabled modules within same package.
    # The two modules shouldn't be mixed up in one bucket.
    items = [
        11,
        13,
        8885,
        8887,
        8889,
        21,
        23,
        9995,
        9997,
        9999,
    ]
    items_copy = list(items)

    for i in range(5):
        _shuffle_items(items, bucket_key=modulus_2_key, disable=disable_if_gt_1000)
        if items != items_copy:
            assert items[items.index(8885) : items.index(8885) + 3] == [8885, 8887, 8889]
            assert items[items.index(9995) : items.index(9995) + 3] == [9995, 9997, 9999]
            return

    assert False


def test_shuffle_respects_seed():
    sorted_items = list(range(30))

    for seed in range(20):
        # Reset
        items1 = list(range(30))
        _shuffle_items(items1, seed=seed)

        assert items1 != sorted_items

        items2 = list(range(30))
        _shuffle_items(items2, seed=seed)

        assert items2 == items1



import collections

import pytest

pytest_plugins = "pytester"


Call = collections.namedtuple("Call", field_names=("package", "module", "cls", "name"))


def _get_test_calls(result):
    """
    Returns a tuple of test calls in the order they were made.
    """
    calls = []

    for c in result.reprec.getcalls("pytest_runtest_call"):
        calls.append(
            Call(
                package=c.item.module.__package__,
                module=c.item.module.__name__,
                cls=(c.item.module.__name__, c.item.cls.__name__) if c.item.cls else None,
                name=c.item.name,
            )
        )
    return tuple(calls)


@pytest.fixture
def get_test_calls():
    """
    Returns a function to get runtest calls out from testdir.pytestrun result object.
    """
    return _get_test_calls


@pytest.fixture
def twenty_tests():
    code = []
    for i in range(20):
        code.append("def test_a{0}(): assert True\n".format(str(i).zfill(2)))
    return "".join(code)


@pytest.fixture
def twenty_cls_tests():
    code = []
    for i in range(20):
        code.append("\tdef test_b{0}(self): self.assertTrue\n".format(str(i).zfill(2)))
    return "".join(code)



# -*- coding: utf-8 -*-

import random
from collections import OrderedDict, namedtuple

from random_order.cache import FAILED_FIRST_LAST_FAILED_BUCKET_KEY

"""
`bucket` is a string representing the bucket in which the item falls based on user's chosen
bucket type.

`disabled` is either a falsey value to mark that the item is ready for shuffling (shuffling is not disabled),
or a truthy value in which case the item won't be shuffled among other items with the same key.

In some cases it is important for the `disabled` to be more than just True in order
to preserve a distinct disabled sub-bucket within a larger bucket and not mix it up with another
disabled sub-bucket of the same larger bucket.
"""
ItemKey = namedtuple("ItemKey", field_names=("bucket", "disabled", "x"))
ItemKey.__new__.__defaults__ = (None, None)


def _shuffle_items(items, bucket_key=None, disable=None, seed=None, session=None):
    """
    Shuffles a list of `items` in place.

    If `bucket_key` is None, items are shuffled across the entire list.

    `bucket_key` is an optional function called for each item in `items` to
    calculate the key of bucket in which the item falls.

    Bucket defines the boundaries across which items will not
    be shuffled.

    `disable` is a function that takes an item and returns a falsey value
    if this item is ok to be shuffled. It returns a truthy value otherwise and
    the truthy value is used as part of the item's key when determining the bucket
    it belongs to.
    """

    if seed is not None:
        random.seed(seed)

    # If `bucket_key` is falsey, shuffle is global.
    if not bucket_key and not disable:
        random.shuffle(items)
        return

    def get_full_bucket_key(item):
        assert bucket_key or disable
        if bucket_key and disable:
            return ItemKey(bucket=bucket_key(item, session), disabled=disable(item, session))
        elif disable:
            return ItemKey(disabled=disable(item, session))
        else:
            return ItemKey(bucket=bucket_key(item, session))

    # For a sequence of items A1, A2, B1, B2, C1, C2,
    # where key(A1) == key(A2) == key(C1) == key(C2),
    # items A1, A2, C1, and C2 will end up in the same bucket.
    buckets = OrderedDict()
    for item in items:
        full_bucket_key = get_full_bucket_key(item)
        if full_bucket_key not in buckets:
            buckets[full_bucket_key] = []
        buckets[full_bucket_key].append(item)

    # Shuffle inside a bucket

    bucket_keys = list(buckets.keys())

    for full_bucket_key in buckets.keys():
        if full_bucket_key.bucket == FAILED_FIRST_LAST_FAILED_BUCKET_KEY:
            # Do not shuffle the last failed bucket
            continue

        if not full_bucket_key.disabled:
            random.shuffle(buckets[full_bucket_key])

    # Shuffle buckets

    # Only the first bucket can be FAILED_FIRST_LAST_FAILED_BUCKET_KEY
    if bucket_keys and bucket_keys[0].bucket == FAILED_FIRST_LAST_FAILED_BUCKET_KEY:
        new_bucket_keys = list(buckets.keys())[1:]
        random.shuffle(new_bucket_keys)
        new_bucket_keys.insert(0, bucket_keys[0])
    else:
        new_bucket_keys = list(buckets.keys())
        random.shuffle(new_bucket_keys)

    items[:] = [item for bk in new_bucket_keys for item in buckets[bk]]
    return


def _get_set_of_item_ids(items):
    return set(item.nodeid for item in items)


def _disable(item, session):
    if hasattr(item, "get_closest_marker"):
        marker = item.get_closest_marker("random_order")
    else:
        marker = item.get_marker("random_order")
    if marker:
        is_disabled = marker.kwargs.get("disabled", False)
        if is_disabled:
            # A test item can only be disabled in its parent context -- where it is part of some order.
            # We use parent name as the key so that all children of the same parent get the same disabled key.
            return item.parent.name
    return False



import pytest


class XdistHooks:
    def pytest_configure_node(self, node: pytest.Item) -> None:
        seed = node.config.getoption("random_order_seed")
        node.workerinput["random_order_seed"] = seed



# -*- coding: utf-8 -*-
import py
import pytest


@pytest.fixture
def tmp_tree_of_tests(testdir):
    """
    Creates a directory structure:
        tmpdir/
            foo.py

    """

    utils_package = testdir.mkpydir("utils")
    utils_package.join("__init__.py").write("")

    utils_package.join("foo.py").write(
        py.code.Source('''
        def add(a, b):
            """
            >>> add(1, 1)
            2
            >>> add(0, 0)
            0
            """
            return a + b

        def subtract(a, b):
            """
            >>> subtract(1, 1)
            0
            >>> subtract(0, 2)
            -2
            """
            return a - b
    ''')
    )

    return testdir


@pytest.mark.parametrize(
    "bucket",
    [
        "global",
        "package",
        "module",
        "class",
        "parent",
        "grandparent",
        "none",
    ],
)
def test_doctests(tmp_tree_of_tests, get_test_calls, bucket):
    result1 = tmp_tree_of_tests.runpytest(
        "--doctest-modules",
        "--random-order-bucket={0}".format(bucket),
        "--verbose",
        "-s",
    )
    result1.assert_outcomes(passed=2, failed=0)
    assert "PytestWarning" not in result1.stdout.str()
    assert "PytestWarning" not in result1.stderr.str()



import pytest


@pytest.mark.parametrize("disabled", [True, False])
def test_marker_disables_random_order_in_module(testdir, twenty_tests, get_test_calls, disabled):
    testdir.makepyfile(
        "import pytest\n" + ("pytestmark = pytest.mark.random_order(disabled={0})\n".format(disabled)) + twenty_tests
    )

    result = testdir.runpytest("--random-order", "-v")
    result.assert_outcomes(passed=20)
    names = [c.name for c in get_test_calls(result)]
    sorted_names = sorted(list(names))

    if disabled:
        assert names == sorted_names
    else:
        assert names != sorted_names


@pytest.mark.parametrize("disabled", [True, False])
def test_marker_disables_random_order_in_class(testdir, twenty_cls_tests, get_test_calls, disabled):
    testdir.makepyfile(
        "import pytest\n\n"
        + "from unittest import TestCase\n\n"
        + "class MyTest(TestCase):\n"
        + "\tpytestmark = pytest.mark.random_order(disabled={0})\n".format(disabled)
        + twenty_cls_tests
        + "\n"
    )

    result = testdir.runpytest("--random-order", "-v")
    result.assert_outcomes(passed=20)
    names = [c.name for c in get_test_calls(result)]
    sorted_names = sorted(list(names))

    if disabled:
        assert names == sorted_names
    else:
        assert names != sorted_names



def test_help_message(testdir):
    result = testdir.runpytest(
        "--help",
    )
    result.stdout.fnmatch_lines(
        [
            "pytest-random-order options:",
            "*--random-order-bucket={global,package,module,class,parent,grandparent,none}*",
            "*--random-order-seed=*",
        ]
    )


def test_markers_message(testdir):
    result = testdir.runpytest(
        "--markers",
    )
    result.stdout.fnmatch_lines(
        [
            "*@pytest.mark.random_order(disabled=True): disable reordering*",
        ]
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding="utf-8").read()


setup(
    name="pytest-random-order",
    author="Jazeps Basko",
    author_email="jazeps.basko@gmail.com",
    maintainer="Jazeps Basko",
    maintainer_email="jazeps.basko@gmail.com",
    license="MIT",
    url="https://github.com/jbasko/pytest-random-order",
    description="Randomise the order in which pytest tests are run with some control over the randomness",
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    packages=[
        "random_order",
    ],
    include_package_data=True,
    python_requires=">=3.9",
    install_requires=[
        "pytest",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Pytest",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Testing",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    keywords="pytest random test order shuffle",
    entry_points={
        "pytest11": [
            "random_order = random_order.plugin",
        ],
    },
)



import pytest


def acceptably_failing_shuffle_items(items, **kwargs):
    # Does not mess up items collection
    raise ValueError("shuffling failed")


def critically_failing_shuffle_items(items, **kwargs):
    # Messes up items collection, an item is effectively lost
    items[1] = items[0]
    raise ValueError("shuffling failed")


def critically_not_failing_shuffle_items(items, **kwargs):
    # This shuffler doesn't raise an exception but it does lose test cases
    items[1] = items[0]


@pytest.fixture
def simple_testdir(testdir):
    testdir.makepyfile("""
        def test_a1():
            assert True

        def test_a2():
            assert True
    """)
    return testdir


def test_faulty_shuffle_that_preserves_items_does_not_fail_test_run(monkeypatch, simple_testdir):
    monkeypatch.setattr("random_order.plugin._shuffle_items", acceptably_failing_shuffle_items)

    result = simple_testdir.runpytest("--random-order")
    result.assert_outcomes(passed=2)
    result.stdout.fnmatch_lines("""
        *pytest-random-order plugin has failed with ValueError*
    """)


def test_faulty_shuffle_that_loses_items_fails_test_run(monkeypatch, simple_testdir):
    monkeypatch.setattr("random_order.plugin._shuffle_items", critically_failing_shuffle_items)
    result = simple_testdir.runpytest("--random-order")
    result.assert_outcomes(passed=0, failed=0, skipped=0)
    result.stdout.fnmatch_lines("""
        *INTERNALERROR> RuntimeError: pytest-random-order plugin has failed with ValueError*
    """)


def test_seemingly_ok_shuffle_that_loses_items_fails_test_run(monkeypatch, simple_testdir):
    monkeypatch.setattr("random_order.plugin._shuffle_items", critically_not_failing_shuffle_items)
    result = simple_testdir.runpytest("--random-order")
    result.assert_outcomes(passed=0, failed=0, skipped=0)
    result.stdout.fnmatch_lines("""
        *INTERNALERROR> RuntimeError: pytest-random-order plugin has failed miserably*
    """)



class Config:
    @classmethod
    def default_value(cls, value):
        return "default:" + str(value)

    def __init__(self, config):
        self._config = config

    @property
    def bucket_type(self):
        if not self.is_enabled:
            return "none"
        else:
            return self._remove_default_prefix(self._config.getoption("random_order_bucket"))

    @property
    def is_enabled(self):
        return self._config.getoption("random_order_enabled") or any(
            not self._config.getoption(name).startswith("default:")
            for name in ("random_order_bucket", "random_order_seed")
        )

    @property
    def seed(self):
        return self._remove_default_prefix(self._config.getoption("random_order_seed"))

    def _remove_default_prefix(self, value):
        if value.startswith("default:"):
            return value[len("default:") :]
        return value
